

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import argparse
import sys
import os
import time
from datetime import timedelta

import tensorflow as tf

import numpy as np

FLAGS = None
from .config import DATASET_DIR, TRAIN_FILE, VALIDATION_FILE, MODEL_NAME, VGG_WEIGHTS_FILE, HEIGHT_BINS, NUM_EPOCHS, BATCH_SIZE, SUMMARIES_DIR, TRAIN_KEEP_PROB
from .models import ConvnetVGG16

def _parse_function(example_proto, augmentation=False):
    """Extract image, labels out of example_proto"""
    int64Feat = lambda val: tf.FixedLenFeature((), tf.int64, default_value=val)
    floatFeat = lambda val: tf.FixedLenFeature((), tf.float32, default_value=val)
    features = {"image_encoded": tf.FixedLenFeature((), tf.string, default_value=""),
                            "width": int64Feat(480),
                            "height":int64Feat(480),
                            "depth":int64Feat(3),
                            "height_reading": floatFeat(0),
                            "rain_detection": int64Feat(0),
                            "rain_quantity": floatFeat(0.),
                            "sun_detection": int64Feat(0),
                            "sun_insolation": floatFeat(0.)}
    parsed_features = tf.parse_single_example(example_proto, features)
    image = tf.image.decode_png(parsed_features['image_encoded'], 3)
    image = tf.to_float(image)
    height_reading = tf.to_float(parsed_features['height_reading'])
    # ignore parsed_features['rain_detection'] and parsed_features['sun_detection']
    rain_quantity = parsed_features['rain_quantity']
    sun_insolation = parsed_features['sun_insolation']
    if augmentation:
        # filp at random for data variety
        image = tf.image.random_flip_up_down(tf.image.random_flip_left_right(image))

    return image, height_reading, rain_quantity, sun_insolation

def heights_to_labels_np(heights):
    """Heights to bin index or if is infinity: index = HEIGHT_BINS.size + 1 """
    inf_idx = num_height_classes() - 1
    # (len(HEIGHT_BINS) + 1) height classes + 1 infinity class
    return np.where(np.isinf(heights), inf_idx, np.digitize(heights, HEIGHT_BINS))

def heights_to_labels(heights):
    """Heights to bin index or if is infinity: index = HEIGHT_BINS.size + 1 """
    bin_indecies = tf.py_func(heights_to_labels_np, [heights], tf.int64)
    return tf.reshape(bin_indecies, [-1])


def num_height_classes():
    """(len(HEIGHT_BINS) + 1) height classes + 1 infinity class"""
    return len(HEIGHT_BINS) + 2


def heights_rains_to_labels(heights, rains):
    h_labels = heights_to_labels(heights)
    # 10 height classes + 1 raining class
    hr_labels = tf.py_func(lambda rn, lbl: np.where(rn > 0, 10, lbl), [rains, h_labels], tf.int64)
    return hr_labels


def print_val_info(step, train_loss, accuracy, trained_batches, start_time):
    #print time and progress info
    sec = time.time() - start_time
    batches_per_sec = trained_batches/sec
    print('step {}, training loss {:g}, AC:{:.2f} ({:.2f} batches/sec) [{!s}]'.format(\
        step, train_loss, accuracy, batches_per_sec, timedelta(seconds=int(sec))))


def checkpoint_restore(sess, saver, checkpoint_dir, norestore):
    ckpt = tf.train.get_checkpoint_state(checkpoint_dir)
    if ckpt and not norestore and ckpt.model_checkpoint_path:
        print('restoring model from:', ckpt.model_checkpoint_path)
        saver.restore(sess, ckpt.model_checkpoint_path)
        return True
    return False


def dataset_iterators(train_filename, validation_filename, batch_size, augmentation):
    parse_func = lambda e: _parse_function(e, augmentation)
    # make iterator for training data
    train_dataset = tf.data.TFRecordDataset(train_filename).map(parse_func, num_parallel_calls=3)
    train_dataset = train_dataset.batch(batch_size)
    training_iterator = train_dataset.make_initializable_iterator()
    # make iterator for validation data
    val_dataset = tf.data.TFRecordDataset(validation_filename).map(parse_func, num_parallel_calls=2)
    val_dataset = val_dataset.batch(batch_size)
    validation_iterator = val_dataset.make_initializable_iterator()

    #create handle for feeding one dataset for each run
    handle = tf.placeholder(tf.string, shape=[])
    iterator = tf.data.Iterator.from_string_handle(handle, train_dataset.output_types, train_dataset.output_shapes)

    return iterator, handle, training_iterator, validation_iterator


def input_data(iterator):
    #read images and labels
    images_big, height_readings, rain_quantity, sun_insolation = iterator.get_next()
    labels = heights_to_labels(height_readings)
    # resize for vgg16 (224x224)
    images = tf.image.resize_images(images_big, (224,224))

    #infrequent summary of input data
    tf.summary.image('cloud_image', images, max_outputs=1, collections=['infrequent'])
    hist_heights = tf.where(tf.is_inf(height_readings), -100 * tf.ones_like(height_readings), height_readings)
    tf.summary.histogram('height_readings', hist_heights, collections=['infrequent'])

    return images, labels, height_readings, rain_quantity, sun_insolation


def unique_sum_dir(summary_dir, step, count=1):
    """Find a summary directory that is not in use already"""
    summary_path = '{}-{}'.format(summary_dir, step)
    if count > 1:
        summary_path = '{}_{}'.format(summary_path, count)

    if not os.path.isdir(summary_path):
        return summary_path
    return unique_sum_dir(summary_dir, step, count+1)

### MODEL TRAINING


def main(_):
    #file locations
    train_summary_location = os.path.join(FLAGS.summaries_dir, 'train')
    val_summary_location = os.path.join(FLAGS.summaries_dir, 'validation')

    model_save_filename= os.path.join(FLAGS.summaries_dir, MODEL_NAME)
    train_filename = os.path.join(DATASET_DIR, TRAIN_FILE)
    validation_filename = os.path.join(DATASET_DIR, VALIDATION_FILE)

    vgg_weight_file = os.path.join(DATASET_DIR, VGG_WEIGHTS_FILE)
    # conf
    validation_after_steps = 5
    train_keep_prob = TRAIN_KEEP_PROB if FLAGS.dropout else 1.0
    with tf.name_scope('input'):
        # use train and validation datasets
        iterator, handle, training_iterator, validation_iterator = \
            dataset_iterators(train_filename, validation_filename, FLAGS.batch_size, FLAGS.augment)

        images, labels, height_readings, rain_quantity, sun_insolation = \
            input_data(iterator)


    # Build the graph for the conv net
    conv_net = ConvnetVGG16(images, classes=num_height_classes())
    keep_prob = conv_net.keep_prob

    #conv_losses = conv_net.loss_classes_with_inf_and_reg()
    conv_losses = conv_net.cross_entropy_loss(labels)
    conv_loss = tf.reduce_mean(conv_losses)
    conv_accuracy = conv_net.accuracy(labels)

    with tf.name_scope('predictions'):
        class_preds = conv_net.class_predictions()
        tf.summary.histogram('classes', class_preds)

    with tf.name_scope('adam_optimizer'):
        global_step = tf.Variable(0, name='global_step', trainable=False)
        train_step = tf.train.AdamOptimizer(1e-4).minimize(conv_loss, global_step=global_step)

    # initialize variables
    init_iterators = tf.group(training_iterator.initializer, validation_iterator.initializer)
    init_op = tf.group(tf.global_variables_initializer(), init_iterators)

    # merge summaries
    merged = tf.summary.merge_all()
    infreq_merged = tf.summary.merge_all(key='infrequent')

    saver = tf.train.Saver()
    w_saver = tf.train.Saver(var_list=conv_net.all_vars)
    # start session
    with tf.Session() as sess:

        # restore from checkpoint if possible (and not disabled)
        norestore = FLAGS.norestore and not FLAGS.init_weights_dir
        if checkpoint_restore(sess, saver, FLAGS.summaries_dir, norestore):
            # iterators have to be initialized (epoch reset)
            sess.run(init_iterators)
            step = sess.run(global_step)
            print('Continue training:')
            # summary directory specific for this run
            train_summary_location = unique_sum_dir(train_summary_location, step)
            val_summary_location = unique_sum_dir(val_summary_location, step)
        elif (os.path.isdir(train_summary_location) or os.path.isdir(val_summary_location)):
            # do not start without restoring if summaries already exist (protection)
            raise Exception("Summary location already exists! Delete, change or restore from directory.")
        elif FLAGS.init_weights_dir:
            w_ckpt = tf.train.get_checkpoint_state(FLAGS.init_weights_dir)
            if w_ckpt and w_ckpt.model_checkpoint_path:
                sess.run(init_op)
                print('Loading weigthts from checkpoint:', w_ckpt.model_checkpoint_path)
                w_saver.restore(sess, w_ckpt.model_checkpoint_path)
            else:
                raise Exception("The initial weight loading checkpoint '{}' not found!".format(FLAGS.init_weights_dir))
        else:
            sess.run(init_op)
            #load pretrained weights of vgg16
            conv_net.load_weights(vgg_weight_file, sess)
            print('Begin training:')

        print('num classes:', conv_net.num_height_classes)
        print('dropout_prob:', train_keep_prob)
        print('augmentation:', FLAGS.augment)
        print('batch_size:', FLAGS.batch_size)
        print('epochs:', FLAGS.num_epochs)
        print('train/validation dataset:', train_filename, validation_filename)
        # iterator handles
        training_handle = sess.run(training_iterator.string_handle())
        validation_handle = sess.run(validation_iterator.string_handle())

        # summary writers
        train_writer = tf.summary.FileWriter(train_summary_location, sess.graph)
        validation_writer = tf.summary.FileWriter(val_summary_location)

        epoch = 0
        start_time = time.time()
        trained_batches = 0
        try:
            while epoch < FLAGS.num_epochs:
                    step = sess.run(global_step)
                    try:
                        #training step
                        summary, _ = sess.run([merged, train_step], feed_dict={keep_prob: train_keep_prob, handle: training_handle})
                        trained_batches += 1
                        if step % 2 == 0:
                            train_writer.add_summary(summary, step)

                    except tf.errors.OutOfRangeError:
                        # one epoch is done
                        epoch += 1
                        # save model during training
                        if not FLAGS.nosave and epoch < FLAGS.num_epochs:
                            print('Saving model after epoch to:', model_save_filename)
                            saver.save(sess, model_save_filename, global_step=global_step)
                        #reinitialize training iterator (starting new epoch)
                        sess.run([training_iterator.initializer])
                        continue

                    if step % validation_after_steps == 0:
                        #validation step
                        try:
                            summary, infreq_summary, train_loss, accuracy = \
                                sess.run([merged, infreq_merged, conv_loss, conv_accuracy],\
                                 feed_dict={keep_prob: 1.0, handle: validation_handle})

                            # write to summary
                            validation_writer.add_summary(summary, step)
                            # infrequent summary
                            if step % (5*validation_after_steps) == 0:
                                validation_writer.add_summary(infreq_summary, step)
                            print_val_info(step, train_loss, accuracy, trained_batches, start_time)
                        except tf.errors.OutOfRangeError:
                            #reinitialize the validation_iterator
                            sess.run([validation_iterator.initializer])

        except (KeyboardInterrupt, SystemExit):
            print('Interrupted training after %d epochs, %d total steps.' % (epoch, step))
        else:
            print('Done training for %d epochs, %d total steps.' % (epoch, step))

        # save model afterwards
        if not FLAGS.nosave:
            print('Saving model to:', model_save_filename)
            saver.save(sess, model_save_filename, global_step=global_step)



if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--norestore', action='store_true')
    parser.add_argument('--init_weights_dir', type=str,
                        default=None,
                        help='Provide checkpoint restore directory for intial weights (otherwise normal initialization)')
    parser.add_argument('--augment', action='store_true')
    parser.add_argument('--dropout', action='store_true')
    parser.add_argument('--nosave', action='store_true')
    parser.add_argument('--num_epochs', type=int,
                        default=NUM_EPOCHS,
                        help='Number of epochs for training')
    parser.add_argument('--batch_size', type=int,
                        default=BATCH_SIZE,
                        help='Batch size for learning')
    parser.add_argument('--summaries_dir', type=str,
                        default=SUMMARIES_DIR,
                        help='Directory for storing summary')

    FLAGS, unparsed = parser.parse_known_args()
    tf.app.run(main=main, argv=[sys.argv[0]] + unparsed)


